package tapl6

import combinatorParsing2._

/** This object implements a parser for Featherweight Java
 *  check <a href="http://citeseer.ist.psu.edu/igarashi99featherweight.html">
 *  the paper</a>.
 */
object FJ extends TokenParsers with TestSuite  {

  val lexical: StdLexical = new StdLexical
  import lexical._
  import implicitCoercions._
  
  reserved.incl("class", "extends", "super", "return", "new")
  specialChars.incl('.', '{', '}', ',', '=', ';')
  
  /** <pre>
   *  Program     ::= { ClassDef } Expr
   * </pre>
   */
  def Prog: Parser[Tree] = positioned(
      rep(ClsDef) ++ Expr                      ^^ pair2fun2(Program)
  )
            
  /** <pre>
   *  ClassDef ::= "class" C extends C "{" 
   *                   { FieldDef } ConstructorDef  { MethodDef } 
   *               "}"
   *  </pre>
   */
  def ClsDef: Parser[ClassDef] = positioned(
      "class" +++ ident ++ "extends" ++ ident ++ "{" ++
        rep(FldDef) ++ CtorDef ++ rep(MethDef) ++ 
      "}"    ^^ tuple52fun5(ClassDef)
    | failure("illegal start of class definition"))

  /** <pre>
   *  FieldDef       ::= ident ident ";"
   *  </pre>
   */
  def FldDef: Parser[FieldDef] = positioned(
	ident ++ ident +++ ";"                       ^^ pair2fun2(FieldDef)
    | failure("illegal start of field definition"))

  /** <pre>
   *  ConstructorDef ::= ident "(" ArgList ")" "{"
   *                       "super" "(" Arguments ")" ";"
   *                       { ident "." ident "=" ident }
   *                     "}"
   *  </pre>
   */
  def CtorDef: Parser[ConstructorDef] = positioned(      
      ident ++ "(" ++ ParamList ++ ")" ++ "{" ++
        "super" ++ "(" ++ VarList ++ ")" ++ ";" ++
        rep(Init) ++
      "}"  ^^ tuple42fun4(ConstructorDef)
    | failure("illegal start of constructor"))

  /**
   * Initializer ::= ident "." ident "=" ident ";"
   */
  def Init: Parser[Initializer] = positioned(
      ident +++ "." ++ ident ++ "=" ++ ident ++ ";" ^^ triple2fun3({(obj, field, rhs) => Initializer(obj, field, Var(rhs)) })
  )
    
  /** <pre>
   *  ParamList ::= ident ident { "," ident ident }
   *            | empty
   *  </pre>
   */
  def ParamList: Parser[List[FieldDef]] = (
        ident ++ ident ++ rep("," ++ ident ++ ident  ^^ pair2fun2(FieldDef)) ^^ triple2fun3( { (tpe, name, fields) => FieldDef(tpe, name) :: fields } )
      | success(Nil: List[FieldDef])
      )
      
  /** <pre>
   *  VarList ::= ident ident { "," ident ident }
   *            | epsilon
   *  </pre>
   */
  def VarList: Parser[List[Var]] = (
        ident ++ rep("," ++ ident ^^ Var)   ^^ pair2fun2({ (ident, idents) => Var(ident) :: idents })
      | success(Nil: List[Var])
      )
  
  /** <pre>
   *  MethodDef ::= ident ident "(" Params ")" "{" "return" Expr ";" "}"
   *  </pre>
   */
  def MethDef: Parser[MethodDef] = positioned(
        ident ++ ident ++ "(" +++ ParamList ++ ")" ++ "{" ++ "return" ++ Expr ++ ";" ++ "}" ^^  tuple42fun4(MethodDef)
      | failure("illegal start of method definition")
      )
      
      
   /** <pre>
    *  Expressions ::= [Expr { "," Expr } ]
    *  </pre>
    */
   def Expressions: Parser[List[Expr]] = (
         Expr ++ rep("," ++ Expr)  ^^ pair2fun2({ (e, es) => e :: es })
       | success(Nil: List[Expr])
       )
  /** <pre>
   *  Expr ::= SimpleExpr {"." ident ["(" Expressions ")"]}
   *  </pre>
   */
  def Expr: Parser[Expr] = positioned(
        SimpleExpr ++ rep("." +++ ident ++ opt("(" ++ Expressions ++ ")")) ^^ pair2fun2(&mkExpression)
      )
      
  /** <pre>
   *  SimpleExpr ::= ident
   *               | "new" C "(" Expressions ")"
   *               | "(" C ")" Expr
   *               | "(" Expr ")"
   *  </pre>
   */
  def SimpleExpr: Parser[Expr] = positioned(
        ident                                        ^^ &Var
      | "new" ++ ident ++ "(" ++ Expressions ++ ")"  ^^ pair2fun2(&New)
      | "(" ++ ident ++ ")" ++ Expr                  ^^ pair2fun2(&Cast)
      | "(" ++ Expr ++ ")"
      | failure("illegal start of simple expression")
      )
      
  type Output = Tree
  
  def mkExpression(obj: Expr, rest: List[Pair[String, Option[List[Expr]]]]) = {
    var t: Expr = obj
    def buildPath(xs: List[Pair[String, Option[List[Expr]]]]): Unit = xs match {
      case Pair(meth, Some(args)) :: rest =>
        t = Apply(t, meth, args)
        buildPath(rest)
      case Pair(field, None) :: rest =>
        t = Select(t, field)
        buildPath(rest)
      case Nil => ()
    }
    buildPath(rest)
    t
  }
  
  def parse(in: String): Option[Output] = {
    val s = new lexical.Scanner(in)

    val res = Prog(s)
    res match {
      case Success(tree, _) => Some(tree)
      case Failure(_, _) => Console.println(res); None
    }
  }

  def eval(t: Tree): Tree = {
    PrettyPrinter(t)
    t
  }
  
  def eval(s: String): Option[Tree] = parse(s) match {
    case Some(t) => Some(eval(t))
    case None    => None
  }

  def main(args: Array[String]) {
    if (args.length < 1) {
      Console.println("usage: tapl6.FJ <file>")
      exit(1)
    }
    check(args(0)) 
  }
}
